import numpy as np

class PVM():
    def __init__(self):
        """
        Create the PVM method object
        """
        pass

    def __call__(self, matrix, weights, types, psi = None, phi = None):
        """
        Score alternatives provided in decision matrix `matrix` using criteria `weights` and criteria `types`.

        Parameters
        -----------
            matrix : ndarray
                Decision matrix with m alternatives in rows and n criteria in columns.
            weights: ndarray
                Attribute weights to the criteria. Sum of weights must be equal to 1.
            types: list
                Define the criteria character:
                Motivating criteria are represented by `m`
                Demotivating criteria are represented by `dm`
                Desirable criteria are represented by `d`
                Non-desirable criteria are represented by `nd`
            psi: ndarray
                Preference vector with motivating or desirable values. Providing of psi is optional.
            phi: ndarray
                Preference vector with demotivating or non-desirable values. Providing of psi is optional.

        Returns
        --------
            ndrarray
                Preference values of each alternative. The best alternative has the highest preference value. 

        Examples
        ----------
        >>> pvm = PVM()
        >>> pref = pvm(matrix, weights, types)
        >>> rank = rank_preferences(pref, reverse = True)
        """
        
        return PVM._pvm(matrix, weights, types, psi, phi)

    @staticmethod
    def _pvm(matrix, weights, types, psi, phi):

        # Determining the preference vector
        if (psi is None) or (phi is None):
            psi = np.zeros(matrix.shape[1])
            phi = np.zeros(matrix.shape[1])

            for i in range(len(types)):
                if types[i] == 'm':
                    psi[i] = np.quantile(matrix[:,i], 0.75)
                    phi[i] = np.quantile(matrix[:,i], 0.25)
                elif types[i] == 'dm':
                    psi[i] = np.quantile(matrix[:,i], 0.25)
                    phi[i] = np.quantile(matrix[:,i], 0.75)
                elif types[i] == 'd':
                    psi[i] = np.max(matrix[:, i])
                    phi[i] = np.min(matrix[:, i])
                elif types[i] == 'nd':
                    psi[i] = np.min(matrix[:, i])
                    phi[i] = np.max(matrix[:, i])
                else:
                    raise ValueError('Only `m`, `dm`, `d` and `nd` criteria types are accepted.')

        # Normalization of decision matrix
        nmatrix = matrix / np.sqrt(np.sum(np.square(matrix), axis = 0))

        # Normalization of preference vector
        psi_prim = psi / np.sqrt(np.sum(np.square(matrix), axis = 0))
        phi_prim = phi / np.sqrt(np.sum(np.square(matrix), axis = 0))

        T_prim = psi_prim - phi_prim
        
        T_prim_prim = T_prim / np.sqrt(np.sum(np.square(T_prim)))

        # Determination of importance factor `mi`
        mi_v = np.zeros(matrix.shape[0])
        mi_d = np.zeros(matrix.shape[0])
        mi_nd = np.zeros(matrix.shape[0])
        l_mi_v, l_mi_d, l_mi_nd = 0, 0, 0
        
        for i in range(len(types)):
            if types[i] == 'm' or types[i] == 'dm':
                mi_v += (nmatrix[:,i] - phi_prim[i]) * T_prim_prim[i] * weights[i]
                l_mi_v += 1
            elif types[i] == 'd':
                mi_d += weights[i]**2 * (nmatrix[:,i] - psi_prim[i])**2
                l_mi_d += 1
            elif types[i] == 'nd':
                mi_nd += weights[i]**2 * (nmatrix[:,i] - psi_prim[i])**2
                l_mi_nd += 1

        mi_d = np.sqrt(mi_d)
        mi_nd = np.sqrt(mi_nd)

        mi = (mi_v * l_mi_v - mi_d * l_mi_d + mi_nd * l_mi_nd) / (l_mi_v + l_mi_d + l_mi_nd)

        return mi


import numpy as np
from .mcda_method import MCDA_method

class VMCM(MCDA_method):
    def __init__(self):
        """
        Create the VMCM method object.
        """

        pass


    def _elimination(self, matrix):
        """
        Calculate significance coefficient values for each criterion.
        Criteria with significance coefficient values between 0 and 0.1 are recommended to be eliminated from the considered criteria set.

        Parameters
        --------------
            matrix : ndarray
                Decision matrix with m alternatives in rows and n criteria in columns.

        Examples
        --------------
        >>> vmcm = VMCM()
        >>> vmcm._elimination(matrix)
        """

        # Elimination of variables
        v = np.std(matrix, axis = 0, ddof = 1) / np.mean(matrix, axis = 0)
        to_eliminate = []
        print('Elimination of variables stage (significance coefficient of features):')
        for el, vj in enumerate(v):
            print(f'C{el + 1} = {vj:.4f}')
            if 0 <= vj <= 0.1:
                to_eliminate.append('C' + str(el + 1))
        print('Criteria to eliminate:')
        if not(to_eliminate):
            print('None')
        for te in to_eliminate:
            print(te)



    def _weighting(self, matrix):
        """
        Calculate criteria weights

        Parameters
        -------------
            matrix : ndarray
                Decision matrix with m alternatives in rows and n criteria in columns.

        Returns
        -------------
            ndarray
                Vector with criteria weights

        Examples
        ------------
        >>> vmcm = VMCM()
        >>> weights = vmcm._weighting(matrix)
        """

        v = np.std(matrix, axis = 0, ddof = 1) / np.mean(matrix, axis = 0)
        return v / np.sum(v)
    

    def _normalization(self, matrix):
        """
        Calculates normalized matrix

        Parameters
        --------------
            matrix : ndarray
                Decision matrix with m alternatives in rows and n criteria in columns.

        Returns
        ---------------
            ndarray
                Normalized matrix

        Examples
        ---------------
        >>> vmcm = VMCM()
        >>> norm_matrix = vmcm._normalization(matrix)
        """

        norm_matrix = (matrix - np.mean(matrix, axis = 0)) / np.std(matrix, axis = 0, ddof = 1)

        return norm_matrix
    

    def _pattern_determination(self, matrix, types):
        """
        Automatic determination of pattern and anti-pattern

        Parameters
        --------------
            matrix : ndarray
                Decision matrix with m alternatives in rows and n criteria in columns.

        Returns
        --------------
            ndarray, ndarray
                Two vectors including values respectively of pattern and anti-pattern

        Examples
        --------------
        >>> vmcm = VMCM()
        >>> pattern, antipattern = vmcm._pattern_determination(matrix, types)
        """

        # Normalization of variables
        norm_matrix = self._normalization(matrix)

        # Determination of pattern and anti-pattern
        q1 = np.quantile(norm_matrix, 0.25, axis = 0)
        q3 = np.quantile(norm_matrix, 0.75, axis = 0)

        pattern = np.zeros(matrix.shape[1])
        anti_pattern = np.zeros(matrix.shape[1])

        pattern[types == 1] = q3[types == 1]
        pattern[types == -1] = q1[types == -1]

        anti_pattern[types == 1] = q1[types == 1]
        anti_pattern[types == -1] = q3[types == -1]

        return pattern, anti_pattern
    

    def _classification(self, m):
        """
        Assign evaluated objects to classes

        Parameters
        -------------
            m : ndarray
                Vector with values of synthetic measure

        Returns
        -------------
            ndarray
                Vector including classes assigned to evaluated objects

        Examples
        --------------
        >>> vmcm = VMCM()
        >>> pref = vmcm(matrix, weights, types)
        >>> classes = vmcm._classification(pref)
        """

        # Classification of objects
        m_mean = np.mean(m)
        m_std = np.std(m, ddof = 1)

        cl = np.zeros(len(m))

        cl[m >= m_mean + m_std] = 1
        cl[np.logical_and(m >= m_mean, m < m_mean + m_std)] = 2
        cl[np.logical_and(m >= m_mean - m_std, m < m_mean)] = 3
        cl[m < m_mean - m_std] = 4

        return cl


    def __call__(self, matrix, weights, types, pattern, anti_pattern):
        """
        Score alternatives provided in decision matrix `matrix` with m alternatives in rows and 
        n criteria in columns using criteria `weights` and criteria `types`.

        Parameters
        ----------
            matrix : ndarray
                Decision matrix with m alternatives in rows and n criteria in columns.
            weights: ndarray
                Vector with criteria weights. Sum of weights must be equal to 1.
            types : ndarray
                Vector with criteria types. Profit criteria are represented by 1 and cost by -1.
            pattern : ndarray
                Vector with values of pattern
            anti_pattern : ndarray
                Vector with values of anti-pattern

        Returns
        -------
            ndrarray
                Vector with preference values of each alternative. The best alternative has the highest preference value. 

        Examples
        ---------
        >>> vmcm = VMCM()
        >>> pattern, antipattern = vmcm._pattern_determination(matrix, types)
        >>> pref = vmcm(matrix, weights, types, pattern, antipattern)
        >>> rank = rank_preferences(pref, reverse = True)
        """

        VMCM._verify_input_data(matrix, weights, types)
        return VMCM._vmcm(self, matrix, weights, types, pattern, anti_pattern)


    @staticmethod
    def _vmcm(self, matrix, weights, types, pattern, anti_pattern):
        # Normalization of variables
        norm_matrix = self._normalization(matrix)

        # Weighting of normalized matrix
        weighted_matrix = norm_matrix * weights

        # Construction of the synthetic measure
        m = np.sum((weighted_matrix - anti_pattern) * (pattern - anti_pattern), axis = 1) / np.sum((pattern - anti_pattern)**2)

        return m